(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 *
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;
open Logic.Operators;;

(*** MODULES ******************************************************************)
module Co = Complexity;;
module F = Format;;
module Fun = Function;;
module H = Hashtbl;;
module RewriteM = Rewritingx.Monad;;
module P = Problem;;
module Sig = Signature;;
module Proj = Filtering;;

(*** TYPES ********************************************************************)
type context = {
  af_p : (Fun.t * int, Logic.p) H.t;
  arith_w : Logic.arith;
  arith_m : Logic.arith;
  weight_a : (Fun.t * int, Logic.a) H.t;
  state : Sig.t;
  num_steps : int;
};;

type flags = {
  help : bool ref;
  recursive : bool ref;
  mulex : bool ref;
  mbits : int ref;
  wbits : int ref;
  defs : bool ref;
  solver : Logic.solver ref;
  nsteps : int ref;
};;

type t = {
  sp : Proj.t;
  input : P.t;
  output : P.t;
  generalized : bool;
};;

(*** GLOBALS ******************************************************************)
let code = "sc";;
let name = "Subterm Criterion Processor";;
let comment = "Applies the subterm criterion processor.";;
let keywords = ["simple projection"; "subterm criterion"; "termination"];;

let flags = {
  help = ref false;
  recursive = ref false;
  mulex = ref false;
  mbits = ref 2;
  wbits = ref 2;
  defs = ref false;
  solver = ref Logic.MiniSat;
  nsteps = ref 0;
};;

let spec =
  let help = "Prints information about flags." in
  let spec = [
    ("--help", Arg.Set flags.help, help);
    ("-help", Arg.Set flags.help, help);
    ("-h", Arg.Set flags.help, help);
    ("-sat", Arg.Unit (fun () -> flags.solver := Logic.MiniSat), "Uses SAT backend (default).");
    (*("-sat", Arg.Set flags.sat, "Uses SAT backend (default).");*)
    ("-smt", Arg.Unit (fun () -> flags.solver := Logic.Yices), "Uses Yices backend (default).");
    (*("-smt", Arg.Clear flags.sat, "Uses SMT backend.");*)
    ("-rec", Arg.Set flags.recursive, "Allow recursive simple projections.");
    ("-mulex", Arg.Set flags.mulex, "Allow projections to multisets of terms).");
    ("-mbits", Arg.Set_int flags.mbits,
      "Bits used for multiplicity of terms in multisets corresponding to left- and right-hand sides (default 2).");
    ("-wbits", Arg.Set_int flags.wbits,
      "Bits used for multiplicity (weight) of arguments in projections (default 2).");
    ("-defs", Arg.Set flags.defs,
      "Allow projection of defined symbols (only relevant for '-rec' and '-mulex'; default \"false\").");
    ("-solver",
      Arg.String (fun raw_flags ->
        let tool::args = String.split ~d:"_" raw_flags in
        flags.solver := Logic.ExternalSmt (tool, args)),
      "Use external SMT solver (expects tool-name and flags separated by underscores).");
    ("-nsteps", Arg.Set_int flags.nsteps, "Number of rewrite steps before checking for subterms (default 0).");
  ] in
  Arg.alignx 80 spec
;;

let help = (comment, keywords, List.map Triple.drop_snd spec);;

(*** MODULES (part 2) *********************************************************)
module Statex = struct type t = context end;;
module StateM = Util.Monad.Transformer.State (Statex) (Logic.Monad);;
open StateM;;

(*** FUNCTIONS ****************************************************************)
let init () =
  flags.help := false;
  flags.recursive := false;
  flags.mulex := false;
  flags.mbits := 2;
  flags.wbits := 2;
  flags.defs := false;
  flags.solver := Logic.MiniSat;
  flags.nsteps := 0;
;;

(* Constructors and Destructors *)
let make ?(gen = false) sp input output = {
  sp = sp;
  input = input;
  output = output;
  generalized = gen;
};;

let get_ip p = p.input;;
let get_op p = p.output;;

(* Complexity Bounds *)
let complexity c _ = Co.mul c Co.other;;

(* Compare Functions *)
let equal p q = P.equal p.input q.input && P.equal p.output q.output;;

(* Printers *)
let (>>=) = RewriteM.(>>=);;

let fprintf fs fmt p =
  F.fprintf fmt "@[<1>%s:@\n@[<1>simple projection:@\n" name;
  Proj.fprintfm fmt p.sp >>= fun _ ->
  F.fprintf fmt "@]@\n@[<1>problem:@\n";
  P.fprintfm fmt p.output >>= fun _ ->
  F.fprintf fmt "@]@\n"; List.hd fs fmt >>= fun _ ->
  Monad.return (F.fprintf fmt "@]")
;;

let fprintfx_mset_proj_entry e =
  let pos p = XmlOutput.int "position" (p + 1) in
  (match e with Proj.Collapsing p -> [p] | Proj.List ps -> ps)
  |> List.map pos
  |> XmlOutput.node "status"
;;

let fprintfx_mset_proj af = XmlOutput.node "multisetArgumentFilter" (List.map (fun (f, fi) ->
  XmlOutput.node "multisetArgumentFilterEntry" [
    RewriteM.fprintfx_fun f;
    (fun fmt -> RewriteM.find_ari f >>= fun i -> XmlOutput.int "arity" i fmt);
    fprintfx_mset_proj_entry fi
  ]
) (Proj.to_list af))
;;

let fprintfx fs p =
  let proj =
    if p.generalized then fprintfx_mset_proj p.sp
    else Proj.fprintfx p.sp
  in
  let output =
    if p.generalized then
      (*give only removed DPs*)
      let dps = Trs.diff (P.get_dps p.input) (P.get_dps p.output) in
      XmlOutput.node "dps" [Trs.fprintfx dps]
    else P.fprintfx p.output
  in
  XmlOutput.node "subtermProc" [proj; output; List.hd fs]
;;

(* general functions *)
let (>>=) = StateM.(>>=);;

(* functions lifted from Logic into StateM *)
let fresh_bool = lift Logic.fresh_bool;;
let fresh_arith a = get >>= fun c -> lift (Logic.fresh_arith a);;
let eval_p p ass = p >>= fun p -> lift (Logic.eval_p p ass);;
let eval_a a ass = a >>= fun a -> lift (Logic.eval_a a ass);;

let neg = liftM Logic.neg;;
let ($->$) a b = liftM2 (<->>) a b;;
let ($|$) a b = liftM2 (<|>) a b;;
let ($&$) a b = liftM2 (<&>) a b;;
let ($?$) a b = liftM2 (<?>) a b;;
let ($:$) a b = liftM2 (<:>) a b;;
let ($*$) a b = liftM2 (<*>) a b;;
let ($=$) a b = liftM2 (<=>) a b;;
let ($>=$) a b = liftM2 (<>=>) a b;;
let ($>$) a b = liftM2 (<>>) a b;;

let map_op op f ls = sequence (List.map f ls) >>= (return <.> op);;
let mapi_op op f ls = sequence (List.mapi f ls) >>= (return <.> op);;
let gen_op op f n = sequence (List.gen f n) >>= (return <.> op);;
let map_and f = map_op Logic.big_and f;;
let gen_and f = gen_op Logic.big_and f;;
let map_or f = map_op Logic.big_or f;;
let mapi_or f = mapi_op Logic.big_or f;;
let gen_or f = gen_op Logic.big_or f;;
let gen_sum f = gen_op Logic.big_sum f;;

(* actual content starts here *)
let context arith_m arith_w nsteps state = {
  af_p = H.create 512;
  arith_m = arith_m;
  arith_w = arith_w;
  weight_a = H.create 512;
  state = state;
  num_steps = nsteps;
};;

(* administrative functions *)
let cache_m tbl f k =
  if H.mem tbl k then return (H.find tbl k)
  else (f k >>= fun v -> (H.add tbl k v; return v))
;;

(* subterm criterion encoding starts here *)
(* caching variables *)
let pos f i = get >>= fun c -> cache_m c.af_p (const fresh_bool) (f, i);;
let weight f i = get >>= fun c -> cache_m c.weight_a (const (fresh_arith c.arith_w)) (f, i);;
let arity f = get >>= fun c -> return (Sig.find_ari f c.state);;

(* encoding of standard subterm criterion *)
let encode_rule cmp rule = match Rule.to_terms rule with
  | (Term.Var _), _ (* fallthrough *)
  | _, (Term.Var _) -> failwith "encode_rule: variable left- or right-hand side"
  | Term.Fun (f, ts), Term.Fun (g, ss) when f = g ->
    mapi_or (fun i (ti, si) -> cmp ti si $&$ pos f i) (List.combine ts ss)
  | Term.Fun (f, ts), Term.Fun (g, ss) ->
    mapi_or (fun j sj ->
      (mapi_or (fun i ti -> cmp ti sj $&$ pos f i) ts) $&$ pos g j)
      ss
;;

let top = return Logic.top;;
let zero = return Logic.zero;;
let one = return Logic.one;;
let plog = function true -> Logic.top | false -> Logic.bot;;
let pst s t = return (plog (Term.is_proper_subterm t s));;
let st s t  = return (plog (Term.is_subterm t s));;

let collapsing f a i = gen_and (fun j -> if i = j then pos f i else neg (pos f j)) a;;

let encode_sp = map_and (fun f -> arity f >>= fun n -> gen_or (collapsing f n) n);;

let rewr_pst trs s t =
  get >>= fun c ->
  if c.num_steps <= 0 then pst s t
  else
    (Rewrite.reducts ~n:c.num_steps s trs
    |> List.tl
    |> map_or (fun u -> st u t)) $|$ pst s t
;;

let rewr_st trs s t =
  get >>= fun c ->
  if c.num_steps <= 0 then st s t
  else
    (Rewrite.reducts ~n:c.num_steps s trs
    |> map_or (fun u -> st u t))
;;

let encode s w =
  let rs = Trs.to_list s in
  map_or (encode_rule (rewr_pst w)) rs $&$
  map_and (encode_rule (rewr_st w)) rs $&$
  encode_sp (Trs.roots s)
;;

(*try to use propositional variable "proj f" that is true
whenever "f" is projected at all*)
let encode_no_proj f = arity f >>= gen_and (neg <.> pos f);;

(* encoding of subterm criterion with recursive projections *)
let encode_rule_rec cmp_subterms rule =
  let cmp s t = cmp_subterms s t $&$
    (List.map Term.root [s; t]
    |> List.map_option id
    |> map_and encode_no_proj)
  in
  let rec encr s t = match t with
    | Term.Fun (g, ts) -> cmp s t $|$ mapi_or (fun j tj -> pos g j $&$ encr s tj) ts
    | Term.Var _ -> cmp s t
  in
  let rec encl s t = match s with
    | Term.Fun (f, ss) -> cmp s t $|$ mapi_or (fun i si -> pos f i $&$ encl si t) ss
    | Term.Var _ -> cmp s t
  in
  let rec enc s t = match (s, t) with
    | (Term.Fun (f, ss), Term.Fun (g, ts)) ->
      if f = g then cmp s t $|$ mapi_or (fun i (si, ti) -> pos f i $&$ enc si ti) (List.combine ss ts)
      else
        cmp s t $|$
          mapi_or (fun i si -> pos f i $&$ mapi_or (fun j tj -> pos g j $&$ enc si tj) ts) ss $|$
          mapi_or (fun j tj -> pos g j $&$ encr s tj) ts $|$
          mapi_or (fun i si -> pos f i $&$ encl si t) ss
    | (Term.Fun (f, ss), Term.Var _) -> cmp s t $|$ mapi_or (fun i si -> pos f i $&$ enc si t) ss
    | (Term.Var _, Term.Fun (g, ts)) -> mapi_or (fun j tj -> pos g j $&$ enc s tj) ts
    | (Term.Var _, Term.Var _) -> cmp s t
  in Rule.to_terms rule |> uncurry enc
;;

let encode_sp_at_most_one =
  map_and (fun f ->
    arity f >>= gen_and (fun j -> gen_and (fun i -> neg (pos f i $&$ pos f j)) j))
;;

(*rules for projected defined symbols have to be oriented weakly*)
let encode_defined_ge enc rule =
  let f = Rule.left_root rule in
  arity f >>= fun n ->
  gen_or (pos f) n $->$ enc rule
;;

let encode_rec defs s w =
  let rs = Trs.to_list s in
  let ss = Trs.to_list w in
  let fs = Trs.funs s @ Trs.funs w in
  map_or (encode_rule_rec pst) rs $&$
  map_and (encode_rule_rec st) rs $&$
  map_and (encode_defined_ge (encode_rule_rec st)) ss $&$
  encode_sp_at_most_one fs $&$
  (if defs then top else map_and encode_no_proj (Trs.left_roots w))
;;

(*given weight w, encode multiplicity of subterm 't' of 's'*)
let rec encode_mult w s t =
  if s = t then
    option encode_no_proj top (Term.root s) $?$ w $:$ zero
  else if Term.is_proper_subterm t s then
    let f = Option.the (Term.root s) in
    arity f >>= gen_sum (fun i ->
      pos f i
        $?$ encode_mult (w $*$ weight f i) (List.nth (Term.args s) i) t
        $:$ zero)
  else zero
;;

let rec filtered_subterms t =
  let root_neq f t = Term.root t |> option (fun g -> g != f) true in
  match t with
  | Term.Fun (f, ts) ->
    List.flat_map filtered_subterms ts
    |> List.filter (root_neq f)
    |> List.insert t
  | Term.Var _ -> [t]
;;

let encode_rule_mulex_neq rule =
  let (l, r) = Rule.to_terms rule in
  (*let ts = List.union (Term.subterms l) (Term.subterms r) in*)
  let ts = List.union (filtered_subterms l) (filtered_subterms r) in
  let msetl = encode_mult one l in
  let msetr = encode_mult one r in
  neg (map_and (fun t -> msetl t $=$ msetr t) ts)
;;

let encode_rule_mulex_ge rule =
  let (l, r) = Rule.to_terms rule in
  (*the finite domain of subterms that have to be considered*)
  (*let ts = List.union (Term.subterms l) (Term.subterms r) in*)
  let ts = List.union (filtered_subterms l) (filtered_subterms r) in
  let msetl = encode_mult one l in
  let msetr = encode_mult one r in
  let upper u =
    List.filter (fun t -> Term.is_proper_subterm u t) ts
    |> map_and (fun t -> msetl t $=$ msetr t)
  in
  map_and (fun t -> upper t $->$ (msetl t $>=$ msetr t)) ts
;;

let encode_mulex defs s w =
  let rs = Trs.to_list s in
  let ss = Trs.to_list w in
  let fs = Trs.funs s @ Trs.funs w in
  map_or encode_rule_mulex_neq rs $&$
  map_and encode_rule_mulex_ge rs $&$
  map_and (encode_defined_ge encode_rule_mulex_ge) ss $&$
  map_and (fun f ->
    arity f >>= gen_and (fun i -> pos f i $->$ (weight f i $>$ zero))) fs $&$
  (if defs then top else map_and encode_no_proj (Trs.left_roots w))
;;

(* decode from assignment *)
let decode_rule gt ass rule = liftM not (eval_p (gt rule) ass);;

let decode_trs gt ass trs =
  StateM.filter (decode_rule gt ass) (Trs.to_list trs) >>= (return <.> Trs.of_list)
;;

let decode_sp_f mulex ass f =
  arity f >>= fun n ->
  sequence (List.gen (fun i ->
    eval_p (pos f i) ass >>= fun b ->
    if mulex && b then eval_a (weight f i) ass >>= fun w -> return (true, (i, Logic.Number.to_int w))
    else if b then return (true, (i, 1))
    else return (false, (i, 0))) n) >>= fun ps ->
  List.filter fst ps
  |> List.map snd
  |> (function
    | [] -> None (*no projection at all*)
    | [(i, j)] -> Some (f, (if j = 1 then Proj.Collapsing i else Proj.List (List.replicate j i)))
    | xs -> Some (f, Proj.List (List.flat_map (fun (i, j) -> List.replicate j i) xs)))
  |> return
;;

let decode_sp mulex ass fs =
  StateM.sequence (List.map (decode_sp_f mulex ass) fs) >>= fun sp ->
  return (Proj.of_list (List.map_option id sp))
;;

let solve signature fs p =
  let configurate s = F.printf "%s@\n%!" s; flags.help := true in
  (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
  if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
  if P.is_dp p then (
    let s = P.get_dps p in
    let w = P.get_trs p in
    let c = context (Logic.nat !(flags.mbits)) (Logic.nat !(flags.wbits)) !(flags.nsteps) signature in
    let gen = !(flags.recursive) || !(flags.mulex) in
    let fs = if gen then Trs.funs s @ Trs.funs w else Trs.roots s in
    let enc =
      if !(flags.recursive) then encode_rec !(flags.defs) s w
      else if !(flags.mulex) then encode_mulex !(flags.defs) s w
      else encode s w
    in
    let gt =
      if !(flags.recursive) then encode_rule_rec pst
      else if !(flags.mulex) then encode_rule_mulex_neq
      else encode_rule (rewr_pst w)
    in
    Logic.run (
      StateM.run c (enc >>= fun phi ->
      StateM.lift (Logic.solve ~solver:!(flags.solver) phi) >>= function
        | None -> return None
        | Some ass ->
          decode_trs gt ass s >>= fun s' ->
          decode_sp !(flags.mulex) ass fs >>= fun sp ->
          return (Some (make sp p (P.set_dps s' p) ~gen:gen))))
  ) else None
;;

(* wrap into state monad *)
let (>>=) = RewriteM.(>>=);;
let solve fs p = RewriteM.get >>= fun s -> RewriteM.return (solve s fs p);;
