(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 *
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module F = Format;;
module P = Problem;;
module XO = XmlOutput;;

(*** TYPES ********************************************************************)
type flags = {
  help : bool ref;
  development : int ref;
  js : bool ref;
  rhs : bool ref;
  remove : int ref;
  reverse : int ref;
 };;

type t = {
 ip : P.t;
 op : P.t;
 pdevelopment : int;
 pjs : bool;
 prhs : bool;
 premove : int;
 preverse : int;
 nrsteps : int;
}

(*** GLOBALS ******************************************************************)
let code = "redundant";;
let name = "Redundant Rules Transformation";;
let keywords = ["redundant";"critical pairs";"confluence"];;
let comment = "Adds or removes redundant (l ->* r for adding, l <->* r for removing) rules.";;
let flags = {
  help = ref false;
  development = ref ~-1;
  js = ref false;
  rhs = ref false;
  remove = ref ~-1;
  reverse = ref ~-1;
};;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.");
  ("-js",Arg.Set flags.js,"Add joining sequences of CPs as rules.");
  ("-development",Arg.Set_int flags.development,"Add rules to make CPs development closed.");
  ("-rhs",Arg.Set flags.rhs,"Add rules by rewriting right-hand sides 1 step.");
  ("-remove",Arg.Set_int flags.remove,"Remove rules whose lhs and rhs are joinable.");
  ("-reverse",Arg.Set_int flags.reverse,"Add reversible rules.");
  ] in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let make p q n = {
 ip = p;
 op = q;
 pdevelopment = !(flags.development);
 pjs = !(flags.js);
 prhs = !(flags.rhs);
 premove = !(flags.remove);
 preverse = !(flags.reverse);
 nrsteps = n;
};;

let (>>=) = Monad.(>>=);;
(* Destructors *)
let get_ip p = p.ip;;
let get_op p = p.op;;

(* Compare Functions *)
let equal p q =
 P.equal (get_ip p) (get_ip q) && P.equal (get_op p) (get_op q)
;;

(* Printers *)
let fprintf fs fmt p =
 F.fprintf fmt "@[<1>%s:@\n" name;
 P.fprintfm fmt (get_op p) >>= fun _ ->
 F.fprintf fmt "@\n"; List.hd fs fmt >>= fun _ -> Monad.return (F.fprintf fmt "@]")
;;

let fprintfx fs p =
  XO.node "redundantRules"
    [P.fprintfx (get_op p); XO.int "nrSteps" p.nrsteps; List.hd fs]
;;

(*** MODULES (part 2) *********************************************************)
(*** FUNCTIONS ****************************************************************)
let init _ =
 flags.help   := false;
 flags.development := ~-1;
 flags.js := false;
 flags.rhs := false;
 flags.remove := ~-1;
 flags.reverse := ~-1;
;;

let seq_to_rule ss = uncurry Rule.of_terms
  (Rewrite.step_get_term (List.hd ss), Rewrite.step_get_reduct (List.last ss))
;;

let solve_dev p =
  let trs = P.get_trs p in
  let n = !(flags.development) in
  let rs =
    P.get_cds p
    |> List.filter (not <.> Diagram.is_development_closed trs)
    |> List.map Diagram.critical_pair
    |> List.filter (fun (s,t) -> List.mem t (Rewrite.reducts ~n:n s trs))
    |> List.map (uncurry Rule.of_terms)
  in
  if List.is_empty rs then
    Monad.return None
  else
    let q = P.make_crp (P.get_language p) (P.get_strategy p)
      (Trs.unique (Trs.union trs (Trs.of_list rs)))
    in
    Monad.return (Some (make p q n))
;;

let solve_rhs p =
  let trs = P.get_trs p in
  let rs =
    Trs.map Rule.to_terms trs
    |> List.flat_map (fun (l, r) -> List.map (Pair.make l) (Rewrite.reducts ~n:1 r trs))
    |> List.map (uncurry Rule.of_terms)
    |> Trs.of_list
  in
  if Trs.size trs = Trs.size rs then
    Monad.return None
  else
    let q = P.make_crp (P.get_language p) (P.get_strategy p) (Trs.unique rs) in
    Monad.return (Some (make p q 2))
;;

let solve_js p =
  let js =
    P.get_cds p
    |> List.flat_map Diagram.joins
    |> List.rev_unzip
    |> uncurry List.append
    |> List.filter (not <.> List.is_empty)
    |> List.filter (not <.> List.is_singleton)
  in
  if List.is_empty js then
    Monad.return None
  else
    let n = List.foldl1 max (List.map List.length js) in
    let jtrs = Trs.of_list (List.map seq_to_rule js) in
    let q = P.make_crp (P.get_language p) (P.get_strategy p)
      (Trs.unique (Trs.union (P.get_trs p) jtrs))
    in
    Monad.return (Some (make p q n))
;;

let rec filter_joinable n trs = function
  | [] -> trs
  | lr :: tl ->
     let rs = Trs.remove lr trs in
     (* tried adding R^-1 for checking joinabolity -- only lost examples *)
     (* let rrs = Trs.filter (Rule.is_rewrite_rule) (Trs.invert rs) in *)
     if Rewrite.are_joinable ~n:n (Rule.lhs lr) (Rule.rhs lr) rs then
       filter_joinable n rs tl
     else
       filter_joinable n trs tl
;;

let solve_rem p =
  let n = !(flags.remove) in
  let trs = P.get_trs p in
  let rs = filter_joinable n trs (Trs.to_list trs) in
  if Trs.size rs = Trs.size trs then
    Monad.return None
  else
    let q = P.make_crp (P.get_language p) (P.get_strategy p) rs in
    Monad.return (Some (make p q n))
;;

let solve_rev p =
  let n = !(flags.reverse) in
  let trs = P.get_trs p in
  let rs =
    Trs.filter (fun lr -> List.mem (Rule.lhs lr) (Rewrite.reducts ~n:n (Rule.rhs lr) trs)) trs
    |> Trs.invert
    |> Trs.filter (Rule.is_rewrite_rule)
  in
  if Trs.is_empty rs then
    Monad.return None
  else
    let q = P.make_crp (P.get_language p) (P.get_strategy p)
      (Trs.unique (Trs.union trs rs))
    in
    Monad.return (Some (make p q n))
;;

let solve fs p =
 let configure s = F.printf "%s@\n%!" s; flags.help := true in
 (try init (); Arg.parsex code spec fs with Arg.Bad s -> configure s);
 if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
 if P.is_ddp p then
   if !(flags.development) >= 0 then
     solve_dev p
   else if !(flags.remove) >= 0 then
     solve_rem p
   else if !(flags.reverse) >= 0 then
     solve_rev p
   else if !(flags.rhs) then
     solve_rhs p
   else (* if !(flags.js) *)
     solve_js p
 else
   Monad.return None
;;
