(*Adaption of Sarah's discrimination tree implementation*)
module Fun = Function;;
module Var = Variable;;
module Pos = Position;;
module M = Monad;;
open Prelude;;
open Util;;

module Make (L : LABEL) =
struct

module Term = Term.Make (L);;
module M = M.Make (L);;
open M;;

exception Empty_pstring

module PString = struct

module Symbol = struct
  type t = F of Fun.t | V;;
  
  let compare = Pervasives.compare;;
  
  let to_string = function
    | V -> "*"
    | F f -> Fun.to_string f
  ;;
  
  let fprintf fmt = function
    | V -> Format.fprintf fmt "@[*@]"
    | F f -> Format.fprintf fmt "@[%s@]" (Fun.to_string f)
  ;;
end

(* for more efficient retrieval operations, tag each symbol in string
   with its after element *)
type label = Symbol.t;;

(* Element(f, next, after) *)
(* these are kind of lists with shortcuts *)
type t = Nil | Element of label * t * t;;

let rec to_string = function
  | Nil -> "[]"
  | Element(e, p, _) -> (Symbol.to_string e) ^ "." ^ (to_string p)
;;

(* the empty pstring *)
let nil = []

(* create pstring from term *)
let rec of_term' next after = function
  | Term.Var _ -> Element(Symbol.V, next, after) (* next=after *)
  | Term.Fun(f, ts) ->
   let next', _ = List.fold_right (fun ti (n, a) ->
     let this = of_term' n a ti in (this, this)) ts (next, after)
   in Element(Symbol.F f, next', after)
;;

let of_term = of_term' Nil Nil 

(* operation to get after element *)
let after = function
  | Nil -> raise Empty_pstring
  | Element(_, _, a)  -> a
;;

let next = function
  | Nil -> raise Empty_pstring
  | Element(_, n, _) -> n
;;

end

module SymbolMap = Map.Partial(PString.Symbol);;

(* the type definition *)
type children = {
  next: node;
  jumps: node list
}
and node =
  | Leaf of int * Term.t list
  | Node of int * (children SymbolMap.t)
;;

type 'a m = 'a M.t
type term = Term.t

type t = node * int

(*** EXCEPTIONS **********************************************************)
exception Malformed_pstring

exception Empty_subtree

exception Not_in_index

(*** FUNCTIONS ***********************************************************)
let next_of n = n.next

let jumps_of n = n.jumps 

let id_of = function
  | Leaf (i, _)
  | Node (i, _) -> i
;;

(********* CONSTRUCT DISCRIMINATION TREE *********************************)

let make_children child jumpchildren = {
  next = child;
  jumps = jumpchildren
};;

let node i = Node (i, SymbolMap.empty);;

let empty = (node 0, 1);;

let make () = empty;;

let get_next_element i p_next table f =
try 
  let children = SymbolMap.find f table in
  (next_of children, jumps_of children, i)
with Not_found -> ( 
  match p_next with
  | PString.Nil -> (Leaf (i, []), [], i+1)
  | _ -> (node i, [], i+1))
;;

let arity = function
  | PString.Symbol.V -> return 0
  | PString.Symbol.F f -> find_ari f
;;

(*TODO: why not use List.drop?*)
let rec drop n xs =
 if n = 0 then xs else 
 match xs with
  | [] -> failwith "empty list in drop"
  | z :: zs -> drop (n-1) zs
;;

(* adds a jumplink j to a list of jump links js. if a link to
   a node with this id is already present, replace *)
let add_nonredundant j js =
  let rec add_nonredundant j acc = function
    | [] -> j :: acc
    | j' :: js when (id_of j) = (id_of j') -> j :: js @ acc
    | j' :: js -> add_nonredundant j (j' :: acc) js
  in
  add_nonredundant j [] js
;;

(* insert pstring into discrimination tree *) 
(* returns (top, after) where top is the topmost node of the extended tree
   and after is a list of possible jump link targets (treated like a 
   stack) *)
let rec insert_pstring k dtree pstring t =
 match dtree, pstring with
(* if leaf reached, simply add value to set *)
  | Leaf(i, ts), PString.Nil -> 
    let leaf = Leaf (i, List.union [t] ts) in return (leaf, [leaf], k)
  | Node(i, n), PString.Element(f, p_next, _) ->
    (* pick appropriate table entry or create new node *)
    let child, jumps, k = get_next_element k p_next n f in
    (* recursively insert remaining pstring *) 
    insert_pstring k child p_next t >>= fun (child', jumpchildren, k) ->
    (* pick appropriate jump target *)
    arity f >>= fun a ->
    let jumpchildren = drop a jumpchildren in
    let jump = List.hd jumpchildren in
    (* create new children for f *)
    let jumps' = add_nonredundant jump jumps in
    let cs = make_children child' jumps' in
    (* update table of current node *)
    let n = SymbolMap.add f cs n in 
    (* return new node n and add n to jump targets *)
    let node = Node(i, n) in
    return (node, node :: jumpchildren, k)
  | _ -> raise Malformed_pstring
;;

(* add entry element into discrimination tree *)
let insert (dtree, k) t =
  let pstring = PString.of_term t in
  insert_pstring k dtree pstring t >>= fun (dtree',_, k) ->
  return (dtree', k)
;;

let size (i:t) = return (snd i)

let is_empty i = size i >>= fun s -> return (s=1)

(********* DESTRUCT DISCRIMINATION TREE *********************************)

let find_children table f =
 try
  let children = SymbolMap.find f table in
  next_of children, jumps_of children
 with
  | Not_found -> raise Not_in_index
;;

let update_jumps (j, exists) js =
 if exists then 
  add_nonredundant j js (* just replace old with updated target *)
 else (* have to remove j from js *)
  let rec remove x acc = function
   | [] -> Format.printf "should not happen\n"; acc
   | y :: xs when (id_of x) = (id_of y) -> List.rev_append xs acc 
   | y :: xs -> remove x (y :: acc) xs
  in
  remove j [] js
;;


(* remove pstring from discrimination tree;
   returns tuple of updated tree and jumpstack, where elements
   on jumpstack are of form (node, b) where the second component
   indicates whether this node still exists *) 
let rec remove_pstring dtree pstring value =
 match dtree, pstring with
  | Leaf(i, set), PString.Nil ->
    if List.mem value set then
     let set' = List.remove value set in
     if [] = set' then 
      return (None, [(dtree, false)])
     else 
      let leaf = Leaf(i, set') in
      return (Some leaf, [(leaf, true)])
    else raise Not_in_index
  | Node(i, table), PString.Element(f, pstring', _) -> (
    (* determine successors *)
    let child, jumps = find_children table f in
    (* recursively remove *)
    remove_pstring child pstring' value >>= fun (treeoption, jumpstack) ->
    (* update jumpstack *)
    arity f >>= fun a ->
    let jumpstack = drop a jumpstack in
    match treeoption with
     | Some child' ->
      let jump = List.hd jumpstack in
      (* update jump list of this node *)
      let jumps' = update_jumps jump jumps in
      (* create new children record *)
      let children' = make_children child' jumps' in
      (* update table of current node *)
      let table = SymbolMap.add f children' table in
      let node = Node(i, table) in
      return (Some node, (node, true) :: jumpstack)
     | None -> (* binding in node for f gets removed *)
      let table = SymbolMap.remove f table in
      if SymbolMap.is_empty table then 
       return (None, (dtree, false) :: jumpstack) (* this node is removed *)
      else
       let node = Node(i, table) in
       return (Some node, (node, true) :: jumpstack) (* return updated node *)
   )
  | _ -> raise Malformed_pstring
;; 

(* removes the value from the index. if not found, Not_in_index is raised *)
let delete (dtree, k) (term, value) =
 let pstring = PString.of_term term in
 remove_pstring dtree pstring value >>= fun (dtreeoption, _) ->
 match dtreeoption with
  | None -> return empty (* DTree got empty *)
  | Some tree -> return (tree, k)
;;

(********* RETRIEVAL OPERATIONS ******************************************)

(***** VARIANTS *****)

let rec retrieve_variants dtree pstring = 
 match dtree, pstring with
  | Leaf(_, set), PString.Nil -> set
  | Node(_,table), PString.Element(f, pstring', _) -> 
   (try
    let dtree, _ = find_children table f in
    retrieve_variants dtree pstring'
   with Not_in_index -> [])
  | _ -> raise Malformed_pstring
;; 

let variant_candidates (dtree, _) term =
 let pstring = PString.of_term term in
 let v = retrieve_variants dtree pstring in
 return v
;;


(***** GENERALIZATIONS *****)

let rec retrieve_generalizations dtree pstring =
 match dtree, pstring with
  | Leaf(_, set), PString.Nil -> set
  | Node(_, table), PString.Element(f, next, after) ->
   List.union 
    (gen_from_node table f next) 
    (gen_from_node table PString.Symbol.V after)
  | _ -> raise Malformed_pstring
and gen_from_node table f pstring =
 try
  let dtree, _ = find_children table f in
  retrieve_generalizations dtree pstring
 with Not_in_index -> []
;;

(* find generalizations for a given term in dtree *)
let generalization_candidates (dtree, _) term =
 let pstring = PString.of_term term in
 let v = retrieve_generalizations dtree pstring in
 return v
;;

(***** UNIFIABLE TERMS *****)

let union f xs =
 let rec union f s = function
  | [] -> s
  | x :: xs -> union f (List.union (f x) s) xs
 in 
 union f [] xs
;;

let rec retrieve_unifiable_terms dtree pstring =
 match dtree, pstring with
  | Leaf(_, set), PString.Nil -> set
  | Node(_, table), PString.Element(f, next, after) ->
   let fmatches = unif_from_node table f next in
   let starmatches = unif_from_node table PString.Symbol.V after in
   if f = PString.Symbol.V then
    let jumps = jumps_from_node table f after in
    List.union fmatches (List.union starmatches jumps)
   else
    List.union fmatches starmatches
  | _, PString.Nil -> failwith "nil" 
  | _ -> raise Malformed_pstring
and unif_from_node table f pstring =
 try
  let dtree, _ = find_children table f in
  retrieve_unifiable_terms dtree pstring
 with Not_in_index -> []
and jumps_from_node table f pstring =
 let acc k v s =
   let _, js = find_children table k in
   let junifs = union (fun n -> retrieve_unifiable_terms n pstring) js in
   List.union s junifs
 in
 SymbolMap.fold acc table [] 
;;

(* find unifiable terms for a given term in dtree *)
let unification_candidates (dtree, _) term =
 let pstring = PString.of_term term in
 let v = retrieve_unifiable_terms dtree pstring in
 return v
;;

end

