(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module Map = Map.Partial (Term);;
module Subst = Substitution;;

(*** TYPES ********************************************************************)
type flags = {
  help : bool ref;
  i : bool ref;
  old : bool ref;
  print : bool ref;
  xml : bool ref;
  gtcap : bool ref;
  nonlinear : bool ref;
};;
type t = Problem.t * bool * Problem.t;;

(*** GLOBALS ******************************************************************)
let code = "edg";;
let name = "EDG Processor";;
let keywords = ["dependency graph";"approximation";"termination"];;
let flags = {
  help = ref false;
  i = ref false;
  old = ref false;
  print = ref false;
  xml = ref false;
  gtcap = ref false;
  nonlinear = ref false
};;

let comment =
 "Remove all edges from the current DG that are not contained in the \
  EDG (approximation of DG based on recursive unification and symmetry)."
;;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.");
  ("-i",Arg.Set flags.i,"Computes the innermost EDG, if possible.");
  ("-old",Arg.Set flags.old,"Removes only edges from l->r to l'->r' where \
   ren(cap(r)) is not unifiable with l' (old edg implementation).");
  ("-p", Arg.Set flags.print, "Only print remaining edges.");
  ("-x", Arg.Set flags.xml, "Use XML-format for output of '-p'.");
  ("-gtcap", Arg.Set flags.gtcap, "Use a general tcap-like non-reachability analysis.");
  ("-nl", Arg.Set flags.nonlinear, "Try to exploit nonlinearity for -gtcap.");
 ] in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let (>>=) = Monad.(>>=);;
let init _ = 
 flags.help := false;
 flags.i := false;
 flags.print := false;
 flags.xml := false;
 flags.gtcap := false;
 flags.nonlinear := false;
;;

(* Destructors *)
let get_ip = Triple.fst;;
let get_op = Triple.thd;;

(* Processor *)

let make_gt_fun trs : Function.t -> Function.t -> bool =
  let module G = Util.Graph.Make (Function) in
  let fs = Trs.funs trs in
  let graph = Trs.to_list trs
  |> List.map Rule.to_terms
  |> List.flat_map (function
    | (Term.Fun (f, _), Term.Fun (g, _)) -> [(f, g)]
    | (Term.Fun (f, _), Term.Var _) -> List.map (Pair.make f) fs
    | _ -> failwith "gt1: variable lhs")
  |> G.of_list
  |> G.transitive_closure
  in
  let gt f g = G.mem_edge (f, g) graph
  in
  gt
;;

(*build a relation "gt" such that "gt s t" whenever "s\sigma" rewrites to
"t\tau" for some "\sigma" and "\tau", based on a combination of tracking the
root-symbols of rewrite rules and a tcap-like analysis.*)
let make_gt gt_fun v tm : Term.t -> Term.t -> bool =
  let gt s t = match (s, t) with
    | (Term.Fun (f, ts), Term.Fun (g, _)) ->
      gt_fun f g &&
      (let s' = Term.Fun (f, List.map (Elogic.tcap' v tm) ts) in
      List.exists (Elogic.ground_matches s') (Elogic.TermMap.find f tm))
    | _ -> true
  in
  gt
;;

(*conclude (general) nonreachability via "root" analysis*)
let nonreach gt (s, t) = match (s, t) with
  | (Term.Fun (f, ss), Term.Fun (g, ts)) -> f <> g && not (gt s t)
  | _ -> false
;;

let var_nonreach gt_fun fs rs =
  let ge f = function
    | Term.Fun (g, _) -> f = g || gt_fun f g
    | Term.Var _ -> true
  in
  let cmp (s, _) (u, _) = compare s u in
  List.filter (Term.is_var <.> fst) rs
  |> List.sort cmp
  |> List.group ~c:cmp
  |> List.exists (fun xts ->
    List.length xts > 1 &&
    List.for_all (fun f -> List.exists (not <.> ge f <.> snd) xts) fs)
;;

(*decompose (general) reachability between terms into several
reachability checks between subterms.*)
let reach_decomp gt =
  let rec rs (s, t) = match (s, t) with
  | (Term.Fun (f, ss), Term.Fun (g, ts)) as st ->
    if f = g && not (gt s t) then List.flat_map rs (List.zip ss ts) else [st]
  | st -> [st]
  in
  rs
;;

let is_arc default_var tm gt_fun gt trs ml mr (x,y) =
  let root = Option.the <.> Term.root <?> "unexpected variable" in
  let s = Rule.lhs x and t = Rule.rhs x and u = Rule.lhs y and v = Rule.rhs y in
  if root t = root u then
    (*TODO: rename could be given as parameter, avoiding many (?) if-then-else*)
    let rename = if !(flags.i) then Rule.rename else Monad.return in
    rename (Rule.of_terms s (Map.find t mr)) >>= fun x' ->
    rename (Rule.of_terms (Map.find u ml) v) >>= fun y' ->
    let s' = Rule.lhs x' and t' = Rule.rhs x' and u' = Rule.lhs y' in
    if !(flags.i) then
      try
        let sub = Elogic.unify t' u and sub' = Elogic.unify u' t in
        let nf sub t = Trs.is_normal_form (Subst.apply_term sub t) trs in
        Monad.return (nf sub s' && nf sub u && nf sub' s)
      with Elogic.Not_unifiable -> Monad.return (false)
    else if !(flags.old) then Monad.return (Elogic.are_unifiable t' u')
    else if !(flags.gtcap) || !(flags.nonlinear) then
      let rs = reach_decomp gt (t, u) in
      Monad.return (
        Elogic.ground_matches_list (List.map (Pair.apply (Elogic.tcap' default_var tm) id) rs) &&
        Elogic.ground_matches u' t &&
        List.for_all (not <.> nonreach gt) rs &&
        (not !(flags.nonlinear) || not (var_nonreach gt_fun (Trs.funs trs) rs))
        (*
        (if Elogic.ground_matches_list (List.map (Pair.apply (Elogic.tcap' default_var tm) id) rs) then true else (Format.printf "tcap'@."; false)) &&
        (if Elogic.ground_matches u' t then true else (Format.printf "rev-tcap@."; false)) &&
        (if List.for_all (not <.> nonreach gt) rs then true else (Format.printf "nonreach@."; false))&&
        (if not (var_nonreach gt_fun (Trs.funs trs) rs) then true else (Format.printf "var-nonreach@."; false))
        *)
      )
    else Monad.return (Elogic.ground_matches t' u && Elogic.ground_matches u' t)
  else Monad.return false
;;

let generate v tm gt_fun gt trs ml mr ns =
 Monad.foldl (fun g n -> Monad.foldl (fun ms m ->
  Monad.liftM (fun c -> if c then m::ms else ms) (is_arc v tm gt_fun gt trs ml mr (n,m))) [] ns >>=
  (Monad.return <.> flip (Graph.add n) g)) Graph.empty ns
;;

let filter_edges trs ml mr g =
 let tm = Elogic.TermMap.of_terms (Trs.lhs trs) in
 let gt_fun = make_gt_fun trs in
 Monad.fresh_var >>= fun x ->
 let v = Term.make_var x in
 let gt = make_gt gt_fun v tm in
 Graph.fold (fun n ms m -> m >>= fun g ->
   let add m ms c = Monad.return (if c then m::ms else ms) in
   Monad.foldl (fun ms m -> is_arc v tm gt_fun gt trs ml mr (n,m) >>= add m ms) [] ms >>=
   (Monad.return <.> flip (Graph.add n) g)) g (Monad.return Graph.empty)
;;

let fprintf_edges fmt g =
  Graph.edges g
  |> Monad.map (fun (r1, r2) ->
    let s = Rule.rhs r1 and t = Rule.lhs r2 in
    Term.fprintfm fmt s >>= fun _ ->
    Format.fprintf fmt " -> ";
    Term.fprintfm fmt t >>= fun _ ->
    Format.fprintf fmt "\n";
    Monad.return ())
;;

let fprintfx_edges fmt g =
  Graph.edges g
  |> Monad.map (fun (r1, r2) ->
    let s = Rule.rhs r1 and t = Rule.lhs r2 in
    XmlOutput.node "rule" [
      XmlOutput.node "lhs" [Term.fprintfx s];
      XmlOutput.node "rhs" [Term.fprintfx t]]
    fmt >>= fun _ ->
    Format.fprintf fmt "\n";
    Monad.return ()
  )
;;

let print_edges xml = if xml then fprintfx_edges else fprintf_edges;;

let solve fs p =
  let configurate s = Format.printf "%s@\n%!" s; flags.help := true in
  (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
  if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
  if Problem.is_dp p then
    let trs = Problem.get_trs p and dps = Problem.get_dps p in
    let g = Problem.the_graph p in
    let (ls,rs) = Pair.map List.unique_hash (Trs.lhs dps,Trs.rhs dps) in
    flags.i := !(flags.i) && Problem.is_it p;
    (* configurate computation *)
    let modify_lhs trs t =
      if !(flags.i) then Trs.tcap t trs
      else if !(flags.old) then Monad.return t
      else Trs.tcap t trs
    in
    let modify_rhs trs t =
      if !(flags.i) then Trs.icap t trs
      else if !(flags.old) then Trs.tcap t trs
      else Trs.tcap t trs
    in
    let add f m t = f t >>= (Monad.return <.> flip (Map.add t) m) in
    Monad.foldl (add (modify_rhs trs)) Map.empty rs >>= fun mr ->
    Monad.foldl (add (modify_lhs (Trs.invert trs))) Map.empty ls >>= fun ml ->
    (* compute arcs *)
    filter_edges trs ml mr (Graph.restrict (Trs.to_list dps) g) >>= fun h ->
    let progress = not (Graph.equal h g) in
    if !(flags.print) then (
      print_edges !(flags.xml) Format.std_formatter h >>= fun _ ->
      Monad.return None
    ) else
      let q = Problem.set_dg (Problem.Partial h) dps trs p in
      Monad.return (if progress then Some (p, !(flags.i), q) else None)
  else Monad.return None
;;

(* Complexity Bounds *)
let complexity c _ = Complexity.mul c Complexity.other;;

(* Compare Functions *)
let equal p q = Problem.equal (get_ip p) (get_ip q);;

(* Printers *)
let fprintf fs fmt p =
 Format.fprintf fmt "@[<1>%s:@\n" name;
 Problem.fprintfm ~g:true fmt (get_op p) >>= fun _ ->
 Format.fprintf fmt "@\n"; List.hd fs fmt >>= fun _ ->
 Monad.return (Format.fprintf fmt "@]")
;;

let fprintfx fs p = XmlOutput.node "depGraphProc" [
  Problem.fprintfx ~g:true (get_op p);
  List.hd fs];;
